@@ -22,6 +22,8 @@
 
 #define LUABIND_BUILDING
 
+#include <boost/foreach.hpp>
+
 #include <luabind/lua_include.hpp>
 
 #include <luabind/config.hpp>
@@ -38,6 +40,24 @@ namespace luabind
 
 namespace luabind { namespace detail {
     
+
+    namespace
+    {
+      struct cast_entry
+      {
+          cast_entry(class_id src, class_id target, cast_function cast)
+            : src(src)
+            , target(target)
+            , cast(cast)
+          {}
+
+          class_id src;
+          class_id target;
+          cast_function cast;
+      };
+
+    } // namespace unnamed
+
     struct class_registration : registration
     {   
         class_registration(char const* name);
@@ -46,40 +66,16 @@ namespace luabind { namespace detail {
 
         const char* m_name;
 
-        // datamembers, some members may be readonly, and
-        // only have a getter function
-        mutable std::map<const char*, detail::class_rep::callback, detail::ltstr> m_getters;
-        mutable std::map<const char*, detail::class_rep::callback, detail::ltstr> m_setters;
-
-        // the operators in lua
-        mutable std::vector<detail::class_rep::operator_callback> m_operators[detail::number_of_operators]; 
         mutable std::map<const char*, int, detail::ltstr> m_static_constants;
 
-        mutable std::vector<class_base::base_desc> m_bases;
-        mutable detail::construct_rep m_constructor;
+        typedef std::pair<type_id, cast_function> base_desc;
+        mutable std::vector<base_desc> m_bases;
 
-        void(*m_destructor)(void*);
-        void(*m_const_holder_destructor)(void*);
-
-        void*(*m_extractor)(void*);
-        const void*(*m_const_extractor)(void*);
-
-        void(*m_const_converter)(void*,void*);
-
-        void(*m_construct_holder)(void*, void*);
-        void(*m_construct_const_holder)(void*, void*);
-
-        void(*m_default_construct_holder)(void*);
-        void(*m_default_construct_const_holder)(void*);
-
-		void(*m_adopt_fun)(void*);
-
-        int m_holder_size;
-        int m_holder_alignment;
-
-        LUABIND_TYPE_INFO m_type;
-        LUABIND_TYPE_INFO m_holder_type;
-        LUABIND_TYPE_INFO m_const_holder_type;
+        type_id m_type;
+        class_id m_id;
+        class_id m_wrapper_id;
+        type_id m_wrapper_type;
+        std::vector<cast_entry> m_casts;
 
         scope m_scope;
         scope m_members;
@@ -116,45 +112,23 @@ namespace luabind { namespace detail {
             m_type
             , m_name
             , L
-            , m_destructor
-            , m_const_holder_destructor
-            , m_holder_type
-            , m_const_holder_type
-            , m_extractor
-            , m_const_extractor
-            , m_const_converter
-            , m_construct_holder
-            , m_construct_const_holder
-            , m_default_construct_holder
-            , m_default_construct_const_holder
-			, m_adopt_fun
-            , m_holder_size
-            , m_holder_alignment);
+		);
 
         // register this new type in the class registry
         r->add_class(m_type, crep);
-        if (!(LUABIND_TYPE_INFO_EQUAL(m_holder_type, LUABIND_INVALID_TYPE_INFO)))
-        {
-            // if we have a held type
-            // we have to register it in the class-table
-            // but only for the base class, if it already
-            // exists, we don't have to register it
-            detail::class_rep* c = r->find_class(m_holder_type);
-            if (c == 0)
-            {
-                r->add_class(m_holder_type, crep);
-                r->add_class(m_const_holder_type, crep);
-            }
-        }
 
-        // constructors
-        m_constructor.swap(crep->m_constructor);
+        lua_pushstring(L, "__luabind_class_map");
+        lua_rawget(L, LUA_REGISTRYINDEX);
+        class_map& classes = *static_cast<class_map*>(
+            lua_touserdata(L, -1));
+        lua_pop(L, 1);
 
-        crep->m_getters.swap(m_getters);
-        crep->m_setters.swap(m_setters);
+        classes.put(m_id, crep);
 
-        for (int i = 0; i < detail::number_of_operators; ++i)
-            crep->m_operators[i].swap(m_operators[i]);
+        bool const has_wrapper = m_wrapper_id != registered_class<null_type>::id;
+
+        if (has_wrapper)
+            classes.put(m_wrapper_id, crep);
 
         crep->m_static_constants.swap(m_static_constants);
 
@@ -169,16 +143,38 @@ namespace luabind { namespace detail {
         m_members.register_(L);
         lua_pop(L, 1);
 
-        for (std::vector<class_base::base_desc>::iterator i = m_bases.begin();
+        lua_pushstring(L, "__luabind_cast_graph");
+        lua_gettable(L, LUA_REGISTRYINDEX);
+        cast_graph* const casts = static_cast<cast_graph*>(
+            lua_touserdata(L, -1));
+        lua_pop(L, 1);
+
+        lua_pushstring(L, "__luabind_class_id_map");
+        lua_gettable(L, LUA_REGISTRYINDEX);
+        class_id_map* const class_ids = static_cast<class_id_map*>(
+            lua_touserdata(L, -1));
+        lua_pop(L, 1);
+
+        class_ids->put(m_id, m_type);
+
+        if (has_wrapper)
+            class_ids->put(m_wrapper_id, m_wrapper_type);
+
+        BOOST_FOREACH(cast_entry const& e, m_casts)
+        {
+            casts->insert(e.src, e.target, e.cast);
+        }
+
+        for (std::vector<base_desc>::iterator i = m_bases.begin();
             i != m_bases.end(); ++i)
         {
             LUABIND_CHECK_STACK(L);
 
             // the baseclass' class_rep structure
-            detail::class_rep* bcrep = registry->find_class(i->type);
+            detail::class_rep* bcrep = registry->find_class(i->first);
 
             detail::class_rep::base_info base;
-            base.pointer_offset = i->ptr_offset;
+            base.pointer_offset = 0;
             base.base = bcrep;
 
             crep->add_base_class(base);
@@ -246,84 +242,18 @@ namespace luabind { namespace detail {
     }
 
     void class_base::init(
-        LUABIND_TYPE_INFO type_id
-        , LUABIND_TYPE_INFO holder_type
-        , LUABIND_TYPE_INFO const_holder_type
-        , void*(*extractor)(void*)
-        , const void*(*const_extractor)(void*)
-        , void(*const_converter_)(void*,void*)
-        , void(*holder_constructor_)(void*,void*)
-        , void(*const_holder_constructor_)(void*,void*)
-        , void(*holder_default_constructor_)(void*)
-        , void(*const_holder_default_constructor_)(void*)
-		, void(*adopt_fun)(void*)
-        , void(*destructor)(void*)
-        , void(*const_holder_destructor)(void*)
-        , int holder_size
-        , int holder_alignment)
+        type_id const& type_id_, class_id id
+      , type_id const& wrapper_type, class_id wrapper_id)
     {
-        m_registration->m_type = type_id;
-        m_registration->m_holder_type = holder_type;
-        m_registration->m_const_holder_type = const_holder_type;
-        m_registration->m_extractor = extractor;
-        m_registration->m_const_extractor = const_extractor;
-        m_registration->m_const_converter = const_converter_;
-        m_registration->m_construct_holder = holder_constructor_;
-        m_registration->m_construct_const_holder = const_holder_constructor_;
-        m_registration->m_default_construct_holder = holder_default_constructor_;
-        m_registration->m_default_construct_const_holder = const_holder_default_constructor_;
-        m_registration->m_destructor = destructor;
-        m_registration->m_const_holder_destructor = const_holder_destructor;
-		m_registration->m_adopt_fun = adopt_fun;
-        m_registration->m_holder_size = holder_size;
-        m_registration->m_holder_alignment = holder_alignment;
+        m_registration->m_type = type_id_;
+        m_registration->m_id = id;
+        m_registration->m_wrapper_type = wrapper_type;
+        m_registration->m_wrapper_id = wrapper_id;
     }
 
-    void class_base::add_getter(
-        const char* name, const boost::function2<int, lua_State*, int>& g)
+    void class_base::add_base(type_id const& base, cast_function cast)
     {
-        detail::class_rep::callback c;
-        c.func = g;
-        c.pointer_offset = 0;
-
-        const char* key = name;
-        m_registration->m_getters[key] = c;
-    }
-
-#ifdef LUABIND_NO_ERROR_CHECKING
-    void class_base::add_setter(
-        const char* name
-        , const boost::function2<int, lua_State*, int>& s)
-#else
-    void class_base::add_setter(
-        const char* name
-        , const boost::function2<int, lua_State*, int>& s
-        , int (*match)(lua_State*, int)
-        , void (*get_sig_ptr)(lua_State*, std::string&))
-#endif
-    {
-        detail::class_rep::callback c;
-        c.func = s;
-        c.pointer_offset = 0;
-
-#ifndef LUABIND_NO_ERROR_CHECKING
-        c.match = match;
-        c.sig = get_sig_ptr;
-#endif
-
-
-        const char* key = name;
-        m_registration->m_setters[key] = c;
-    }
-
-    void class_base::add_base(const base_desc& b)
-    {
-        m_registration->m_bases.push_back(b);
-    }
-
-    void class_base::add_constructor(const detail::construct_rep::overload_t& o)
-    {
-        m_registration->m_constructor.overloads.push_back(o);
+        m_registration->m_bases.push_back(std::make_pair(base, cast));
     }
 
 	void class_base::add_member(registration* member)
@@ -338,29 +268,6 @@ namespace luabind { namespace detail {
 		m_registration->m_default_members.operator,(scope(ptr));
 	}
 
-#ifndef LUABIND_NO_ERROR_CHECKING
-    void class_base::add_operator(
-        int op_id,  int(*func)(lua_State*), int(*matcher)(lua_State*)
-        , void(*sig)(lua_State*, std::string&), int arity)
-#else
-    void class_base::add_operator(
-        int op_id,  int(*func)(lua_State*)
-        , int(*matcher)(lua_State*), int arity)
-#endif
-    {
-        detail::class_rep::operator_callback o;
-        o.set_fun(func);
-        o.set_match_fun(matcher);
-        o.set_arity(arity);
-
-#ifndef LUABIND_NO_ERROR_CHECKING
-
-        o.set_sig_fun(sig);
-
-#endif
-        m_registration->m_operators[op_id].push_back(o);
-    }
-
     const char* class_base::name() const 
     { 
         return m_registration->m_name; 
@@ -376,17 +283,20 @@ namespace luabind { namespace detail {
         m_registration->m_scope.operator,(s);
     }
 
-	template<class T>
-	void add_custom_name(T i, std::string& s) {}
+    void class_base::add_cast(
+        class_id src, class_id target, cast_function cast)
+    {
+        m_registration->m_casts.push_back(cast_entry(src, target, cast));
+    }
 
-	void add_custom_name(std::type_info const* i, std::string& s)
+	void add_custom_name(type_id const& i, std::string& s)
 	{
 		s += " [";
-		s += i->name();
+		s += i.name();
 		s += "]";
 	}
 
-    std::string get_class_name(lua_State* L, LUABIND_TYPE_INFO i)
+    std::string get_class_name(lua_State* L, type_id const& i)
     {
         std::string ret;
 
@@ -402,19 +312,20 @@ namespace luabind { namespace detail {
         }
         else
         {
-            if (LUABIND_TYPE_INFO_EQUAL(i, crep->holder_type()))
+            /* TODO reimplement this?
+            if (i == crep->holder_type())
             {
                 ret += "smart_ptr<";
                 ret += crep->name();
                 ret += ">";
             }
-            else if (LUABIND_TYPE_INFO_EQUAL(i, crep->const_holder_type()))
+            else if (i == crep->const_holder_type())
             {
                 ret += "smart_ptr<const ";
                 ret += crep->name();
                 ret += ">";
             }
-            else
+            else*/
             {
                 ret += crep->name();
             }
